__author__ = ('Imam Omar Mochtar', ('iomarmochtar@gmail.com',))

import re
from datetime import datetime

class FilterNotFound(Exception):
    pass

class FilterNotMatch(Exception):
    pass


def filter_wrapper(wfilter):
    """
    agar filter bisa mempunyai argumen/parameter
    """
    def callme(**params):
        def injector(*args, **kwargs):
            kwargs['params'] = params
            return wfilter(*args, **kwargs)
        return injector 
    return callme


class Filters(object):

    __register = [
        'NUMBER',
        'DB_DATE',
        'TEXT'
    ]

    def f_number(self, argnum, args, params):
        txt = args[argnum]
        prefix = params.get('prefix', None)
        if prefix:
            ptrn = re.search(r'^\d+(\w)$', txt)

            if ptrn and ptrn.group(1) in prefix:
                txt = txt.replace( ptrn.group(1),  prefix[ptrn.group(1)] )

        if not txt.isdigit():
            raise FilterNotMatch('{} is not a number'.format(txt)) 
        return txt

    def f_text(self, argnum, args, params):
        """
        Join all arguments to the rest of line
        """
        return ' '.join(args[argnum:])

    def f_db_date(self, argnum, args, params):
        tgl = args[argnum]
        date_format = params.get('format', '%Y-%m-%d')

        if params.get('allow_now') and tgl == 'now':
            tgl = datetime.now() 
        else:
            try:
                tgl = datetime.strptime(tgl, date_format)
            except ValueError:
                raise FilterNotMatch('{} is not a valid date'.format(tgl)) 

        return tgl if params.get('convert') else tgl.strftime(date_format)


    def __getattr__(self, attr):
        if attr not in self.__register:
            raise FilterNotFound('Filter with name {} is not found'.format(attr))
        name = attr.lower()
        convert = 'f_{}'.format(name)
        return filter_wrapper( getattr(self, convert) )

filters = Filters()



__author__ = ('Imam Omar Mochtar', ('iomarmochtar@gmail.com',))

"""
Base of cakap library bot class
"""

import logging
import os
import re
import sys
import inspect
import json
from glob import glob
from ConfigParser import ConfigParser
from colorlog import ColoredFormatter
from pprint import pprint
from .utils import Utils 
from .decorators import auth, catch_error, chelp, AutoSet
from .filters import FilterNotMatch
from telegram.ext import Updater, CommandHandler, MessageHandler, Filters, InlineQueryHandler, Job 

# agar bisa jalan async
from telegram.ext.dispatcher import run_async

# untuk job queue
from telegram.ext import Job 


class BotBase(object):

    __metaclass__ = AutoSet

    users = {} 
    allowed_users = []
    allowed_gids = []
    helps = [] 
    debug = False 
    cmd_prefix = 'cmd_' 

    def __init__(self, token, workers=3, users=[], 
        groups=[], name=None, debug=False):

        self.debug = debug
        self.name = name if name else type(self).__name__

        self.allowed_users = users
        self.allowed_gids = groups
        self.init_logger()

        self.updater = Updater(
            token, 
            workers=workers
        )


    def check_auth(self, bot, update, func_name=None):
        """
        Auth & authorization check which using @auth decorator
        """
        chat_id = Utils.chat_id(update)
        if Utils.is_group(update):
            if chat_id not in self.allowed_gids:
                self.logger.warn('receiving message from group id {} is not alowed'.format(chat_id))
                return False 

        user_details = Utils.user_details(update)
        if user_details['username'] not in self.allowed_users:
            self.logger.warn('{} is not allowed for accessing {}'.format(
                    user_details['username'],
                    func_name
                ))
            return False 
        return True 

    def init_logger(self, logger=None, level=logging.DEBUG):
        formatter = ColoredFormatter(
            '%(yellow)s%(asctime)s %(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s',
            # datefmt=None,
            reset=True,
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red',
                }
        )

        handler = logging.StreamHandler()
        handler.setFormatter(formatter)

        if not logger:
            self.logger = logging.getLogger(
                    type(self).__name__ if not self.name else self.name
                )
            logger = self.logger
        logger.addHandler(handler)
        logger.setLevel(level)

    def filter_args(self, argfilters, args):
        """
        Filter given argument parsers
        """
        # if number of filters is more than arguments means not enought
        if len(argfilters) > len(args):
            return 'Need more argument, please check /help for command documentation'

        result = {}
        count = 0
        for argfilter in argfilters:
            name = argfilter['name']
            # if filters didn't exist 
            if not 'filters' in argfilter or not argfilter['filters']:
                result[name] = args[count]
                count += 1
                continue

            for fltr in argfilter['filters']:
                try:
                    result[name] = fltr(argnum=count, args=args)  
                except FilterNotMatch as e:
                    return str(e)
            count += 1 
        
        return result 

    def error_handler(self, msg):
        self.logger.error(msg)

    @chelp('Get available command with it\'s description')
    @auth
    def cmd_help(self, bot, update):
        """Show available help"""
        Utils.send(bot, update, '\n'.join(self.helps))


    @classmethod
    def get_cmd_name(self, name):
        """
        Get name of cmd which is startswith {{cmd_prefix}}
        """
        re_name =  re.search('^{0}(.*?)$'.format(self.cmd_prefix), name)
        return re_name.group(1) if re_name else None 

    def pre_main(self):
        """
        Executed before registering main command handlers
        """
        pass

    def after_main(self):
        """
        Executed after registering main command handlers
        """
        pass

    def polling(self):
        self.updater.start_polling()
        self.updater.idle()

    def main(self):

        self.pre_main()
        for command in inspect.getmembers(self, predicate=inspect.ismethod):
            name = command[0]
            handler = command[1]

            if not name.startswith(self.cmd_prefix):
                continue

            cmd_name = self.get_cmd_name(name)
            cmd_is_async = handler.is_async if hasattr(handler, 'is_async') else False
            cmd_args = handler.args if hasattr(handler, 'args') else None 

            # TODO: help di set dalam bentuk markdown 
            hlp = '/{}'.format(cmd_name)
            if cmd_args:
                hlp += ' - {0}'.format(' '.join( map(lambda x: '[%s]'%x['name'].upper(), cmd_args) )) 

            if hasattr(handler, 'desc'):
                hlp += ' - {0}'.format(handler.desc)

            self.helps.append(hlp)

            self.logger.debug('Registering cmd {}'.format(cmd_name))
           
            self.updater.dispatcher.add_handler(
                    CommandHandler(cmd_name, 
                        run_async(handler) if cmd_is_async else handler, 
                        pass_args=True if cmd_args else False
                    )
                )


        self.after_main()

        self.logger.info('{} is running'.format(self.name))		
        self.polling()



__author__ = ('Imam Omar Mochtar', ('iomarmochtar@gmail.com',))

import subprocess

class Utils(object):

    @staticmethod
    def runCmd(command):
        return subprocess.Popen(
            command, universal_newlines=True, shell=True,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()

    @staticmethod
    def chat_id(update):
        """
        wrapper to get chat id from conversation
        """
        return update.message.chat_id

    @staticmethod
    def user_details(update):
        """
        wrapper to get sender's information 
        """
        return update.message.from_user

    @staticmethod
    def is_group(update):
        """
        wrapper to get to check wether message comes from group or not 
        """
        return update.message.chat.type == 'group' 
        #return update.message.chat.title

    @staticmethod   
    def send(bot, update, text, chat_id=None, reply=False, **kwargs):
        chat_id = chat_id if chat_id else Utils.chat_id(update)
        # set to reply conversation if needed
        if reply:
            kwargs['reply_to_message_id'] = update.message.message_id
        bot.sendMessage(chat_id=chat_id, text=text, **kwargs)

    @staticmethod
    def send_doc(bot, update, path, filename, chat_id=None):
        """
        send document file 
        """
        chat_id = chat_id if chat_id else Utils.chat_id(update)
        bot.sendDocument(chat_id=chat_id, document=open(path, 'rb'), filename=filename)



__author__ = ('Imam Omar Mochtar', ('iomarmochtar@gmail.com',))

"""
Collection of decorators
"""

import inspect
import traceback
from pprint import pprint
from .utils import Utils

def chelp(desc=None, args=[], is_async=False):
    """
    set help and args for cmd
    """
    def getfunc(f):
        def getresult(*args, **kwargs):
            return f(*args, **kwargs)

        getresult.desc = desc 
        getresult.args = args 
        getresult.is_async = is_async

        return getresult 
    return getfunc 

def auth(f):
    def auth_filter(*args, **kwargs):
        obj, bot, update = args[:3]
        f_name = f.__name__
        if not obj.check_auth(bot, update, f_name):
            return None
        return f(*args, **kwargs)
    return auth_filter 

# catch any error
def catch_error(f):
    def ce_filter(*args, **kwargs):
        obj, bot, update = args[:3]
        try:
            # if it's has arguments then do argument filtering
            if hasattr(f, 'args') and f.args:
                clean_args = obj.filter_args(f.args, kwargs.get('args', []))
                if type(clean_args) == str:
                    Utils.send(bot, update, clean_args, reply=True)
                    return
                kwargs['args'] = clean_args

            result = f(*args, **kwargs)
            # if it's returned string means to reply the message
            if type(result) != str:
                return result

            Utils.send(bot, update, result, reply=True)
            
        except Exception as e:
            err_msg = traceback.format_exc()
            msg = err_msg if obj.debug else 'Please see program\'s log for more detail'
            Utils.send(bot, update, 'ERR: %s'%msg, reply=True)
            obj.error_handler(err_msg)
    
    # make sure it's not called multiple times
    for attr in ['desc', 'args', 'is_async']:
        if not hasattr(ce_filter, attr) and hasattr(f, attr):
            setattr(ce_filter, attr, getattr(f, attr)) 

    return ce_filter 

class AutoSet(type):

    def __init__(cls, name, bases, ns):

        prefix = getattr(cls, 'cmd_prefix')

        for name, val in ns.items():
            # if contain cmd_ as it's suffix and function then assign the decorators (with condition)
            if name.startswith(prefix) and inspect.isfunction(val):

                # mandatory decorator to cache error
                setattr(cls, name, catch_error(val))



from setuptools import setup


long_description = ''
# Get the long description from the README file
with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='cakap',
    version='0.4',
    license='GPL',
    long_description=long_description,
    url='https://github.com/iomarmochtar/cakap',
    author='Imam Omar Mochtar',
    author_email='iomarmochtar@gmail.com',
    install_requires=['python-telegram-bot', 'colorlog'],
    packages=['cakap'],
    keywords='Telegram bot framework',
    classifiers=[
    ]
)
